'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; /**
                                                                                                                                                                                                                                                                   * The top-level merge function
                                                                                                                                                                                                                                                                   *
                                                                                                                                                                                                                                                                   * @param {object} remerge map
                                                                                                                                                                                                                                                                   * @return {function} reducer function
                                                                                                                                                                                                                                                                   *
                                                                                                                                                                                                                                                                   */

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var collectionRegex = /^\$(.+)/;

var merge = function merge(map) {
  var debugMode = arguments.length <= 1 || arguments[1] === undefined ? false : arguments[1];


  var _getAccessorKey = function _getAccessorKey(key) {
    // this regex tests if the key is of the form $abcd1234
    var captureAccessor = collectionRegex.exec(key);
    if (captureAccessor) {
      return captureAccessor[1];
    } else {
      return null;
    }
  };

  var _preprocess = function _preprocess(map) {
    var newMap = {};

    for (var key in map) {
      if (key === '_') {
        continue;
      }
      if (newMap.$) {
        (0, _utils.consoleWarning)('More than one collection accessor ' + newMap.$.accessorKeyName);
        continue;
      }

      var isFunction = _lodash2.default.isFunction(map[key]);
      newMap[key.replace(_getAccessorKey(key), "")] = {
        isLeaf: isFunction,
        accessorKeyName: _getAccessorKey(key),
        child: isFunction ? map[key] : _preprocess(map[key])
      };
    }

    return newMap;
  };

  var _initial = function _initial(map) {
    if (!map) {
      return undefined;
    } else if (_lodash2.default.isFunction(map)) {
      return undefined;
    } else if (map['_'] !== undefined) {
      return map['_'];
    }

    var newMap = {};
    for (var key in map) {
      if (!_getAccessorKey(key)) {
        var result = _initial(map[key]);
        if (result !== undefined) {
          newMap[key] = result;
        }
      }
    }
    return Object.keys(newMap).length > 0 ? newMap : null;
  };

  var _process = function _process(_map, state, action) {
    var currentPath = action.type.split('.', 1)[0];
    var newState = _lodash2.default.clone(state);
    var foundPath = false;

    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = Object.keys(_map)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var path = _step.value;

        if (path === currentPath) {
          foundPath = true;
          var _map$path = _map[path];
          var accessorKeyName = _map$path.accessorKeyName;
          var isLeaf = _map$path.isLeaf;
          var child = _map$path.child;


          if (isLeaf) {
            (0, _utils.consoleSuccess)('Executing action at leaf node: ' + currentPath.bold, debugMode);
            return child(newState, action);
          } else {
            var collectionKeyName = child.$ && child.$.accessorKeyName;
            var collectionKey = action && action[collectionKeyName];

            // child is a collection and we should enter because accessor key name is given
            if (collectionKeyName && collectionKey !== undefined) {
              (0, _utils.consoleSuccess)('Navigating collection node: ' + currentPath.bold, debugMode);
              var newCollection = _lodash2.default.clone((0, _utils.getCollectionElement)(newState, path));
              var smallerMap = child.$.child;
              var smallerState = (0, _utils.getCollectionElement)(newCollection, collectionKey);
              var smallerAction = _extends({}, action, {
                type: action.type.split('.').splice(1).join(".")
              });

              var newSmallerState = _process(smallerMap, smallerState, smallerAction);
              (0, _utils.setCollectionElement)(newCollection, collectionKey, newSmallerState);
              (0, _utils.setCollectionElement)(newState, path, newCollection);
            } else {
              (0, _utils.consoleSuccess)('Navigating element node: ' + currentPath.bold, debugMode);
              var smallerMap = child;
              var smallerState = (0, _utils.getCollectionElement)(newState, path);
              var smallerAction = _extends({}, action, {
                type: action.type.split('.').splice(1).join(".")
              });
              var newSmallerState = _process(smallerMap, smallerState, smallerAction);
              (0, _utils.setCollectionElement)(newState, path, newSmallerState);
            }
          }
        }
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    if (!foundPath) {
      (0, _utils.consoleError)('Could not find path: ' + currentPath.bold, debugMode);
    }

    return newState;
  };

  var initialState = _initial(map);
  var computedMap = _preprocess(map);

  return function (state, action) {

    if (action === undefined) {
      (0, _utils.consoleMessage)('Setting up initial state tree', debugMode);
    } else if (!action.type) {
      (0, _utils.consoleError)('Action is missing type', debugMode);
    } else {
      (0, _utils.consoleMessage)(('Received action with type: ' + action.type.bold).underline, debugMode);
    }

    if (state === undefined) {
      return initialState;
    }

    return _process(computedMap, state, action);
  };
};

exports.default = merge;